//----------------------------------------------------------------------------
// union-find framework for random maze generation
// copyright 2010 christopher rasmussen
// university of delaware
// v. 1.02
// * draw() is terrain-aware
// * black-and-white version of draw()
// * added wind terrain

//----------------------------------------------------------------------------

#include "BlockMaze.hh"

//----------------------------------------------------------------------------
//----------------------------------------------------------------------------

BlockMaze::BlockMaze(int rows, int cols)
{
  // make sure maze is going to be big enough

  if (rows < MIN_MAZE_ROWS || cols < MIN_MAZE_COLS) {
    cout << "Must have >= " << MIN_MAZE_ROWS << " rows and " << MIN_MAZE_COLS << " cols\n";
    exit(1);
  }

  num_rows = rows;
  num_cols = cols;

  // initialize all walls 

  vertex.resize(num_rows);

  for (int r = 0; r < num_rows; r++) {
    vertex[r].resize(num_cols);
    for (int c = 0; c < num_cols; c++) {
      vertex[r][c].set_position(r, c);
      vertex[r][c].make_wall();
    }
  }

  compute_wall_counts();

  // initialize underlying set data structure

  S = new DisjointSet(num_rows * num_cols);
}

//----------------------------------------------------------------------------

void BlockMaze::compute_wall_counts()
{
  num_total_walls = num_rows * num_cols;
  max_num_interior_walls = (num_rows - 2) * num_cols - 2 * (num_rows - 2);
  num_exterior_walls = num_total_walls - max_num_interior_walls;

  num_interior_walls = num_empty_spaces = num_fire = num_water = 0;

  for (int r = 1; r < num_rows - 1; r++)
    for (int c = 1; c < num_cols - 1; c++)
      if (vertex[r][c].is_wall())
	num_interior_walls++;
      else {
	num_empty_spaces++;
	if (vertex[r][c].is_water())
	  num_water++;
	else if (vertex[r][c].is_fire())
	  num_fire++;
      }
}

//----------------------------------------------------------------------------

// remove wall at (row, col) and merge cells that it divides, if any

void BlockMaze::knock_down_wall(int row, int col)
{
  int here;

  vertex[row][col].make_empty();

  here = element_index(row, col);

  // merge with N if applicable

  if (!vertex[row - 1][col].is_wall()) 
    S->merge(here, element_index(row - 1, col));

  // merge with S if applicable

  if (!vertex[row + 1][col].is_wall()) 
    S->merge(here, element_index(row + 1, col));

  // merge with E if applicable

  if (!vertex[row][col + 1].is_wall()) 
    S->merge(here, element_index(row, col + 1));

  // merge with W if applicable

  if (!vertex[row][col - 1].is_wall()) 
    S->merge(here, element_index(row, col - 1));

  num_interior_walls--;
  num_empty_spaces++;
}

//----------------------------------------------------------------------------

// how many knock-down-able walls are left?

int BlockMaze::num_ok_to_knock_down()
{
  int total = 0;
  int r, c;

  for (r = 1; r < num_rows - 1; r++)
    for (c = 1; c < num_cols - 1; c++)
      if (vertex[r][c].is_wall() && ok_to_knock_down(r, c))
	total++;

  return total;
}

//----------------------------------------------------------------------------

#define DO_SQUARE_CORNERS

// a wall is ok to knock down if doing so will NOT create a new
// connection between previously-connected spaces.  

bool BlockMaze::ok_to_knock_down(int row, int col)
{
  vector <int> index;
  int i, j;

  // make a list of empty spaces adjacent to wall
  // under consideration...

  index.clear();

  if (!vertex[row - 1][col].is_wall())
    index.push_back(element_index(row - 1, col));
#ifdef DO_SQUARE_CORNERS
  else {
    if (vertex[row][col + 1].is_wall() && !vertex[row - 1][col + 1].is_wall())
      return false;
  }
#endif

  if (!vertex[row + 1][col].is_wall())
    index.push_back(element_index(row + 1, col));
#ifdef DO_SQUARE_CORNERS
  else {
    if (vertex[row][col - 1].is_wall() && !vertex[row + 1][col - 1].is_wall())
      return false;
  }
#endif

  if (!vertex[row][col + 1].is_wall())
    index.push_back(element_index(row, col + 1));
#ifdef DO_SQUARE_CORNERS
  else {
    if (vertex[row + 1][col].is_wall() && !vertex[row + 1][col + 1].is_wall())
      return false;
  }
#endif

  if (!vertex[row][col - 1].is_wall())
    index.push_back(element_index(row, col - 1));
#ifdef DO_SQUARE_CORNERS
  else {
    if (vertex[row - 1][col].is_wall() && !vertex[row - 1][col - 1].is_wall())
      return false;
  }
#endif
  
  // ...and see if any of them are already connected

  for (i = 0; i < index.size(); i++)
    for (j = 0; j < index.size(); j++)
      if (i != j && S->same_equivalence_class(index[i], index[j]))
	return false;

  return true;
}

//----------------------------------------------------------------------------

// brute force approach: pick random cell, check if wall is actually there.  
// repeat until one is found

bool BlockMaze::pick_random_interior_wall(int & row, int & col)
{
  if (num_interior_walls == 0)
    return false;
  else {
    while (1) {
      row = 1 + rand() % (num_rows - 2);
      col = 1 + rand() % (num_cols - 2);
      if (vertex[row][col].is_wall())
	break;
    }
    return true;
  }
}

//----------------------------------------------------------------------------

// brute force approach: pick random cell, check if wall is actually there.  
// repeat until one is NOT found

bool BlockMaze::pick_random_empty_space(int & row, int & col)
{
  // if there are no empty spaces to find, don't bother...

  if (num_interior_walls == max_num_interior_walls)
    return false;
  else {
    while (1) {
      row = 1 + rand() % (num_rows - 2);
      col = 1 + rand() % (num_cols - 2);
      if (!vertex[row][col].is_wall())
	break;
    }
    return true;
  }
}

//----------------------------------------------------------------------------

// turn empty spaces into water or fire until >= water_fraction, >= fire_fraction
// are those kinds of terrain

// assuming compute_wall_counts() has been called, so num_empty_spaces is set

void BlockMaze::add_random_terrain(float water_fraction, float fire_fraction)
{
  // impossible request

  if (water_fraction + fire_fraction > 1.0)
    return;

  float current_water_fraction = (float) num_water / (float) num_empty_spaces;
  float current_fire_fraction = (float) num_fire / (float) num_empty_spaces;
  int row, col;

  while (current_water_fraction < water_fraction) {

    if (!pick_random_empty_space(row, col))
      return;

    if (vertex[row][col].is_water() || vertex[row][col].is_fire())
      continue;

    vertex[row][col].terrain_type = TERRAIN_WATER;
    num_water++;
    current_water_fraction = (float) num_water / (float) num_empty_spaces;
  }

  while (current_fire_fraction < fire_fraction) {

    if (!pick_random_empty_space(row, col))
      return;

    if (vertex[row][col].is_water() || vertex[row][col].is_fire())
      continue;

    vertex[row][col].terrain_type = TERRAIN_FIRE;
    num_fire++;
    current_fire_fraction = (float) num_fire / (float) num_empty_spaces;
  }

}

//----------------------------------------------------------------------------

// print maze including current state of walls

void BlockMaze::print()
{
  int r, c;

  cout << "num interior walls = " << num_interior_walls << "/" << max_num_interior_walls << endl;
  cout << "num exterior walls = " << num_exterior_walls << endl;
  cout << "num equivalence classes = " << S->num_equivalence_classes << endl;

  for (r = 0; r < num_rows; r++) {
    for (c = 0; c < num_cols; c++) {

      if (vertex[r][c].is_wall())
	cout << "#";
      else
	cout << " ";
    }
    cout << "\n";
  }
  cout << "\n";
}

//----------------------------------------------------------------------------

// set all vertex distances to DIST_INFINITY

void BlockMaze::initialize_distances()
{
  for (int r = 0; r < num_rows; r++) 
    for (int c = 0; c < num_cols; c++) {
      vertex[r][c].distance = DIST_INFINITY;
      vertex[r][c].path = NULL;
    }
}

//----------------------------------------------------------------------------

// draw all walls using ncurses

void BlockMaze::draw()
{
  int r, c;

  // color version

  if (has_colors()) {
    for (r = 0; r < num_rows; r++) 
      for (c = 0; c < num_cols; c++) {
	move(r, c);
	if (vertex[r][c].is_wall()) {
	  attron(COLOR_PAIR(WALL_COLOR));  
	  printw(" ");
	  attroff(COLOR_PAIR(WALL_COLOR));
	}
	else if (vertex[r][c].is_water()) {
	  attron(COLOR_PAIR(CYAN_COLOR));  
	  addch(ACS_CKBOARD);
	  attroff(COLOR_PAIR(CYAN_COLOR));
	}
	else if (vertex[r][c].is_fire()) {
	  attron(COLOR_PAIR(RED_COLOR));  
	  addch(ACS_CKBOARD);
	  attroff(COLOR_PAIR(RED_COLOR));
	}
	else if (vertex[r][c].is_wind()) {
	  attron(COLOR_PAIR(PURPLE_COLOR));  
	  addch(ACS_CKBOARD);
	  attroff(COLOR_PAIR(PURPLE_COLOR));
	}
      }
  }

  // black-and-white version

  else {
    for (r = 0; r < num_rows; r++) 
      for (c = 0; c < num_cols; c++) {
	move(r, c);
	if (vertex[r][c].is_wall()) 
	  printw("%%");
	else if (vertex[r][c].is_water()) 
	  printw("w");
	else if (vertex[r][c].is_fire()) 
	  printw("f");
	else if (vertex[r][c].is_wind()) 
	  printw("W");
      }
  }
}

//----------------------------------------------------------------------------

void BlockMaze::draw_distances()
{
  int d;

  for (int r = 0; r < num_rows; r++) 
    for (int c = 0; c < num_cols; c++) {
      move(r, c);
      if (!vertex[r][c].is_wall()) {
	d = vertex[r][c].distance;
	if (d == DIST_INFINITY)
	  printw(".");
	else {
	  if (d < 10)
	    attron(COLOR_PAIR(BLUE_COLOR));
	  else if (d < 20)
	    attron(COLOR_PAIR(GREEN_COLOR));
	  else if (d < 30)
	    attron(COLOR_PAIR(YELLOW_COLOR));
	  else
	    attron(COLOR_PAIR(RED_COLOR));

	  printw("%i", d % 10);

	  if (d < 10)
	    attroff(COLOR_PAIR(BLUE_COLOR));
	  else if (d < 20)
	    attroff(COLOR_PAIR(GREEN_COLOR));
	  else if (d < 30)
	    attroff(COLOR_PAIR(YELLOW_COLOR));
	  else
	    attroff(COLOR_PAIR(RED_COLOR));
	  
	}
      }
    }
}

//----------------------------------------------------------------------------

bool BlockMaze::knock_down_walls_randomly_until_everything_connected(bool interactive)
{
  int r, c;
  bool bottom;

  while (1) {

    // print current state of maze

    //    print();

    // get keyboard input (enter only will do)

    if (interactive)
      cin.get();

    // pick a wall (if there is one) and knock it down.
    // pick_random_interior_wall() seems to get stuck sometimes for very large mazes,
    // even when there are plenty of walls...not sure why

    if (pick_random_interior_wall(r, c) && ok_to_knock_down(r, c))
      knock_down_wall(r, c);
    else 
      continue;
    
    int num_ok = num_ok_to_knock_down();

//     cout << "num ok " << num_ok << endl;
//     print();

    if (num_ok == 0) {
      break;
    }
  }

  // is everything connected?

  if (S->num_equivalence_classes < num_total_walls &&
      S->num_equivalence_classes - num_exterior_walls - num_interior_walls == 1) {
    return true;
  }

  // sometimes it's not fully connected--if so, put walls back up so we can try again

  else {

    for (r = 0; r < num_rows; r++) 
      for (c = 0; c < num_cols; c++) 
	vertex[r][c].make_wall();

    return false;
  }
}

//----------------------------------------------------------------------------
//----------------------------------------------------------------------------
