#ifndef BLOCKMAZE_HH
#define BLOCKMAZE_HH

//----------------------------------------------------------------------------
// union-find framework for random maze generation
// copyright 2010 christopher rasmussen
// university of delaware
// v. 1.01
// * added costs for different terrain types
// * defined constants for terrain types and other .lay elements
// * added add_random_terrain() function
// * added wind terrain

//----------------------------------------------------------------------------

#include "DisjointSet.hh"
#include "Utils.hh"

//----------------------------------------------------------------------------
//----------------------------------------------------------------------------

#define DIST_INFINITY    -1

#define MIN_MAZE_ROWS  5
#define MIN_MAZE_COLS  5

#define TERRAIN_NONE  0
#define TERRAIN_WALL  1
#define TERRAIN_WATER 2
#define TERRAIN_FIRE  3
#define TERRAIN_WIND  4

// time which must be spent in each type of terrain before moving again
// (EMPTY assumed 0)

#define WATER_COST    5
#define FIRE_COST     10
#define WIND_COST     10      // only applies to ghosts

#define LAY_WALL_CHAR         '%'
#define LAY_RUNNER_CHAR       'P'
#define LAY_GHOST_CHAR        'G'
#define LAY_POWER_PILL_CHAR   'o'
#define LAY_WATER_CHAR        'w'
#define LAY_FIRE_CHAR         'f'
#define LAY_WIND_CHAR         'W'

//----------------------------------------------------------------------------

class BlockVertex
{
public:

  BlockVertex() { terrain_type = TERRAIN_WALL; distance = DIST_INFINITY; }
  void set_position(int r, int c) { pos.row = r; pos.col = c; }
  bool is_wall() { return terrain_type == TERRAIN_WALL; }
  bool is_empty() { return terrain_type != TERRAIN_WALL; }  // anything but wall is "empty"
  bool is_none() { return terrain_type == TERRAIN_NONE; }
  bool is_water() { return terrain_type == TERRAIN_WATER; }
  bool is_fire() { return terrain_type == TERRAIN_FIRE; }
  bool is_wind() { return terrain_type == TERRAIN_WIND; }
  void make_wall() { terrain_type = TERRAIN_WALL; }
  void make_empty() { terrain_type = TERRAIN_NONE; }   // NONE is default empty terrain

  Point_2D pos;         // (row, col) of this block
  int terrain_type;     // wall?  empty?  something else?
  int distance;         // used for searching (e.g., BFS)
  BlockVertex *path;    // predecessor vertex in BFS search
  bool known;           // for Dijkstra's algorithm

};

//----------------------------------------------------------------------------

class BlockMaze
{
public:

  BlockMaze(int, int);
  void print();

  bool is_exterior_wall(int row, int col) { 
    return (row == 0 || row == num_rows - 1 || col == 0 || col == num_cols - 1); }

  // convert grid cell to set element index
  int element_index(int row, int col)  { return (col + row * num_cols); }

  void knock_down_wall(int, int);
  bool ok_to_knock_down(int, int);
  int num_ok_to_knock_down();
  bool pick_random_interior_wall(int &, int &);
  bool pick_random_empty_space(int &, int &);
  bool knock_down_walls_randomly_until_everything_connected(bool = false);
  void draw();
  void initialize_distances();
  void draw_distances();
  void compute_wall_counts();
  void add_random_terrain(float, float);

  int num_rows, num_cols;

  int num_interior_walls;
  int max_num_interior_walls;
  int num_exterior_walls;
  int num_total_walls;
  int num_empty_spaces;
  int num_water, num_fire;

  vector < vector <BlockVertex> > vertex;

  DisjointSet *S;
  
};

//----------------------------------------------------------------------------
//----------------------------------------------------------------------------

#endif
