//----------------------------------------------------------------------------
// pacman-like game
// copyright 2010 christopher rasmussen
// university of delaware
// v. 1.02
// * added score/time elapsed drawing
// * added wind terrain

//----------------------------------------------------------------------------

#include "BlockMaze.hh"
#include "Utils.hh"
#include "Agent.hh"

//----------------------------------------------------------------------------
//----------------------------------------------------------------------------

#define DO_SPLASH

int runner_agent_index;

const char QUIT_KEY        = 'e';

const int CONTINUE_PLAYING   = 0;
const int RUNNER_CAUGHT      = 1;
const int LEVEL_CLEARED      = 2;

//----------------------------------------------------------------------------
//----------------------------------------------------------------------------

class MazeRunner
{
public:

  MazeRunner(int rows, int cols);
  MazeRunner(string filename);

  void main_loop();

private:

  int update();
  void draw_splash();
  void draw_gameover(int);
  void draw();
  void draw_power_pills();
  bool handle_key();
  void compute_dimensions_lay(ifstream &, int &, int &);
  void load_lay(string);
  bool check_for_power_pill_eaten();
  bool check_for_level_cleared();

  int num_updates;

  BlockMaze *M;                 // the layout of the environment
  vector <Agent *> agent;
  vector <Point_2D> power_pill;
  GameState GS;

};

//----------------------------------------------------------------------------
//----------------------------------------------------------------------------

// random maze constructor

MazeRunner::MazeRunner(int rows, int cols) 
{ 
  num_updates = 0;

  M = new BlockMaze(rows, cols);   

  // loop until we get a good random maze...haven't seen a failure
  // yet at this level

  int i;
  for (i = 0; i < 5; i++)
    if (M->knock_down_walls_randomly_until_everything_connected())
      break;
  if (i == 5) {
    cout << "Sorry but I failed to generate a maze...your dimensions might be too big\n";
    exit(1);
  }

  // uncomment this if you want random mazes to have terrain cells
  //  M->add_random_terrain(0.1, 0.1);

  // place runner

  runner_agent_index = agent.size();
  Agent *runner = new RunnerAgent("runner", 'c', YELLOW_COLOR);
  M->pick_random_empty_space(runner->position.y, runner->position.x);
  agent.push_back(runner);

  // place power pills

  // at least 1 power pill, plus one more for every 100 squares. 
  // make sure they have some minimal separation and are not on top
  // of runner to start.  not sure if this is impossible for some
  // configurations--haven't seen it yet...
  
  int area = rows * cols;
  int num_power_pills = 1 + area / 100;
  Point_2D p;
  int min_dist, dist;

  for (int i = 0; i < num_power_pills; i++) {
    while (1) {
      M->pick_random_empty_space(p.y, p.x);
      if (p.x == agent[runner_agent_index]->position.x && p.y == agent[runner_agent_index]->position.y)
	continue;
      min_dist = 100000;
      for (int j = 0; j < power_pill.size(); j++) {
	dist = fabs(p.x - power_pill[j].x) + fabs(p.y - power_pill[j].y);
	if (dist < min_dist)
	  min_dist = dist;
      }
      if (min_dist >= 5)
	break;
    }
    power_pill.push_back(p);
  }
}

//----------------------------------------------------------------------------

// load from .lay constructor

MazeRunner::MazeRunner(string filename)
{
  num_updates = 0;

  M = new BlockMaze(MIN_MAZE_ROWS, MIN_MAZE_COLS);
  
  load_lay(filename);
  M->compute_wall_counts();
  // don't change terrain
}

//----------------------------------------------------------------------------

// render current state of everything

void MazeRunner::draw()
{
  int i;

  // clear screen

  erase();

  // draw maze itself

  M->draw();

  // "score" info

  move(M->num_rows, 0);
  printw("%i turns elapsed", num_updates);
  move(M->num_rows + 1, 0);
  printw("%i pills to go", power_pill.size());
  
  // draw agents

  for (i = 0; i < agent.size(); i++) 
    agent[i]->draw_debug(M);

  draw_power_pills();

  for (i = 0; i < agent.size(); i++) 
    agent[i]->draw(M);

  // make drawing visible

  refresh();
}

//----------------------------------------------------------------------------

void MazeRunner::draw_power_pills()
{
  attron(COLOR_PAIR(WHITE_COLOR));  

  for (int i = 0; i < power_pill.size(); i++) {
    move(power_pill[i].y, power_pill[i].x);
    printw("o");
  }

  attroff(COLOR_PAIR(WHITE_COLOR));  
}

//----------------------------------------------------------------------------

// if the runner touched a power pill, it's gone

// is runner agent position coincident with power pill position?
// if so, remove it and return true
// else return false

bool MazeRunner::check_for_power_pill_eaten()
{
  for (int i = 0; i < power_pill.size(); i++)
    if (power_pill[i].x == agent[runner_agent_index]->position.x &&
	power_pill[i].y == agent[runner_agent_index]->position.y) {
      power_pill.erase(power_pill.begin() + i);
      return true;
    }

  return false;
}

//----------------------------------------------------------------------------

// are all power pills gone?

bool MazeRunner::check_for_level_cleared()
{
  return (power_pill.size() == 0);
}

//----------------------------------------------------------------------------

// do computations, etc.

int MazeRunner::update()
{
  num_updates++;

  // pack up game state

  GS.update(agent, power_pill);

  // let agents do their thing

  for (int i = 0; i < agent.size(); i++) {
    agent[i]->update(GS, M);

    // decrement non-zero wait counter, whether or not agent tried to move

    if (agent[i]->terrain_wait_counter > 0)
      agent[i]->terrain_wait_counter--;
  }

  // consequences?

  check_for_power_pill_eaten();

  if (check_for_level_cleared())
    return LEVEL_CLEARED;

  return CONTINUE_PLAYING;
}

//----------------------------------------------------------------------------

// count rows, cols for Berkeley layout file

void MazeRunner::compute_dimensions_lay(ifstream & inStream, int & rows, int & cols)
{
  char c;
  bool got_cols;
  int this_cols;

  rows = cols = 0;
  got_cols = false;
  this_cols = 0;

  while (!inStream.eof()) {

    inStream.get(c);

    if (inStream.eof()) {
      if (got_cols && this_cols > cols / 2)
	rows++;
      break;
    }

    if (c == '\n') {
      rows++;
      got_cols = true;
      this_cols = 0;
    }
    else {
      this_cols++;
      if (!got_cols)
	cols++;
    }

    // echo to console
    
    cout.put(c);  
  }

  cout << endl;

  // rewind
  
  inStream.clear();
  inStream.seekg(0,fstream::beg);
}

//----------------------------------------------------------------------------

// read maze from Berkeley layout file

void MazeRunner::load_lay(string filename)
{
  ifstream inStream;
  char c;
  int rows, cols;

  // try to open file to mutate

  inStream.open(filename.c_str());

  if (inStream.fail()) {
    cout << "Failed to open file " << filename << endl;
    exit(1);
  }

  // how big is grid?

  compute_dimensions_lay(inStream, rows, cols);

  // initialize all walls 

  M->num_rows = rows;
  M->num_cols = cols;

  M->vertex.resize(M->num_rows);

  for (rows = 0; rows < M->num_rows; rows++) {
    M->vertex[rows].resize(M->num_cols);
    for (cols = 0; cols < M->num_cols; cols++)
      M->vertex[rows][cols].set_position(rows, cols);
  }

  // read 'em in now that we know dimensions

  rows = cols = 0;

  while (!inStream.eof()) {

    inStream.get(c);

    if (inStream.eof())
      break;

    if (c == '\n') {
      rows++;
      cols = 0;
    }
    else {

      stringstream out;
      string s;
      Agent *A;

      if (c == LAY_WALL_CHAR)
	M->vertex[rows][cols].make_wall();
      else { 

	M->vertex[rows][cols].make_empty();

	if (c == LAY_RUNNER_CHAR) {
	  runner_agent_index = agent.size();
	  A = new RunnerAgent("runner", 'c', YELLOW_COLOR, cols, rows);
	  agent.push_back(A);
	}
	else if (c == LAY_GHOST_CHAR) {     

	  // NO GHOSTS -- THIS WILL CHANGE FOR PROJECT #3

	}
	else if (c == LAY_POWER_PILL_CHAR)    
	  power_pill.push_back(Point_2D(rows, cols));
	else if (c == LAY_WATER_CHAR)    
	  M->vertex[rows][cols].terrain_type = TERRAIN_WATER;
	else if (c == LAY_FIRE_CHAR)    
	  M->vertex[rows][cols].terrain_type = TERRAIN_FIRE;
	else if (c == LAY_WIND_CHAR)    
	  M->vertex[rows][cols].terrain_type = TERRAIN_WIND;
      }

      cols++;
    }
  }

  // clean up

  inStream.close();
}

//----------------------------------------------------------------------------

// interpret keyboard input...return false if user
// wants to quit

bool MazeRunner::handle_key()
{
  int c;

  c = getch();

  // exit game

  if (c == QUIT_KEY)
    return false;

  // send all other commands to runner

  else
    agent[runner_agent_index]->handle_key(c, M);
    
  return true;
}

//----------------------------------------------------------------------------

// set up drawing, print instructions, and go

void MazeRunner::draw_splash()
{
  // get ready to draw

  init_ncurses();

#ifdef DO_SPLASH

  // instructions

  printw("Welcome to MazeRunner, the wacky game of energy collection!!\n");
  printw("Hit '%c' to exit at any time\n", QUIT_KEY);   
  printw("Ready?\n");   

  refresh();

  // wait on input

  getch();			                                              

#endif

  // need this for animation without user input

  timeout(100);  
}

//----------------------------------------------------------------------------

// last words...

void MazeRunner::draw_gameover(int result)
{
#ifdef DO_SPLASH
  int c;

  timeout(-1);

  erase();

  printw("Goodbye!!\n");

  if (result == RUNNER_CAUGHT) {
    printw("You sure were easy to catch...\n");
    printw("They got you in %i turns\n", num_updates);
    printw("Hit 'e' to exit\n");
  }
  else if (result == LEVEL_CLEARED) {
    printw("Nice running...\n");
    printw("It took you %i turns\n", num_updates);
    printw("Hit 'e' to exit\n");
  }   
  else
    printw("Hit any key to exit\n");   
  
  refresh();

  if (result == RUNNER_CAUGHT || result == LEVEL_CLEARED) {
    do { c = getch(); } while (c != QUIT_KEY);
  }

  // any key will do

  else
    getch();			                                              

#endif

  // clean up

  terminate_ncurses();
}

//----------------------------------------------------------------------------

// initialize, loop until done, and clean up

void MazeRunner::main_loop()
{
  int result;

  // initialize

  draw_splash();

  // main loop

  do {

    result = update();
    draw();

  } while (result == CONTINUE_PLAYING && handle_key());

  // clean up

  draw_gameover(result);
}

//----------------------------------------------------------------------------
//----------------------------------------------------------------------------

int main(int argc, char **argv)
{
  MazeRunner *MR;

  // initialize random number generator

  srand(time(0));

  // initialize

  if (argc == 2) 
    MR = new MazeRunner(argv[1]);
  else if (argc == 3)
    MR = new MazeRunner(atoi(argv[1]), atoi(argv[2]));
  else 
    MR = new MazeRunner(10, 10);

  // run

  MR->main_loop();

  // done

  return 1;
}

//----------------------------------------------------------------------------
//----------------------------------------------------------------------------
